/* unicode.c -- codepage convertion functionality;
 * Copyright (C) 2001-2002, ForestNet IRC Network
 * This is a free software; see file `license' for details.
 * $Id: unicode.c,v 1.29 2002/11/18 12:44:08 hex Exp $
 */

#ifdef _WIN32
# include <io.h>
#else
# include <sys/file.h>
#endif

/* for fstat() */
#include <sys/stat.h>
#include <sys/types.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h> /* strcat */

#include "sys.h" /* MyFree etc */
#include "struct.h"
#include "h.h"
#include "unicode.h"
#include "common.h" /* strncmp for win32 */

#define IsEOL(c) (c == '\r' || c == '\n')
#define IsWS(c) (c == ' ' || c == '\t')

#ifdef _WIN32
# define PATHSEP "\\"
#else
# define PATHSEP "/"
#endif

static CharSet *charsets = NULL;
static TransLitSet *cs_translit = NULL;


/* Transforms a given Unicode character code into an utf-8 sequence, returns
 * a pointer to the first octet of the null-terminated sequence.
 */
__inline static unsigned char *uc2utf(unsigned short uc)
{
	static unsigned char utf8[5] = {0,0,0,0,0};

	/* 1 byte */
	if (uc < 0x80) {
		utf8[3] = (char)uc;
		return &utf8[3];
	}

	/* 2 bytes */
	else if (uc < 0x800) {
		utf8[2] = 0xc0 + ((uc / 64) & 0x1f);
		utf8[3] = 0x80 + (uc & 0x3f);
		return &utf8[2];
	}

	/* 3 bytes */
	else {
		utf8[1] = 0xe0 + ((uc / 4096) & 0x0f);
		utf8[2] = 0x80 + ((uc / 64) & 0x3f);
		utf8[3] = 0x80 + (uc & 0x3f);
		return &utf8[1];
	}
}


__inline static unsigned short utf2uc(unsigned char **p)
{
	unsigned short uc;

	/* 1 byte (what's it doing here) */
	if ((**p & 0x80) == 0x00) {
		uc = **p;
		++(*p);
	}

	/* 2 bytes */
	else if ((**p & 0xe0) == 0xc0 && (*p)[1]) {
		uc = (*((*p)++) & 0x1f) << 6;
		uc |= *((*p)++) & 0x3f;
	}

	/* 3 bytes */
	else if ((**p & 0xf0) == 0xe0 && (*p)[1] && (*p)[2]) {
		uc = (*((*p)++) & 0x1f) << 12;
		uc |= (*((*p)++) & 0x03f) << 6;
		uc |= (*((*p)++) & 0x3f);
	}

	/* unknown length */
	else {
		uc = 0;
		for (++(*p); (**p & 0xc0) == 0x80; ++(*p));
	}

	return uc;
}


static char *read_line(char *buf, size_t size, FILE *in)
{
	char *str, *p;
	size_t idx;

	str = NULL;

	if (!feof(in) && (str = fgets(buf, size, in)) != NULL) {
		if (buf[idx = strcspn(buf, "\r\n")])
			buf[idx] = '\0';
		/* fgets returns truncated string.  if we didn't get
		 * any EOL character, then we couldn't read the whole
		 * string, read the rest now.
		 */
		else while (!feof(in) && !IsEOL(fgetc(in)));

		for (; IsWS(*str); ++str);
		for (p = str; *p != '\0'; ++p)
			*p = tolower(*p);
	}

	return str;
}


static TransLitSet *read_translit(void)
{
	FILE *in;
	u_short uc, ucmin, ucmax;
	u_char linebuf[1024], *str, *p, *map;
	size_t size, sz;
	TransLitSet *cs;

	strcpy(linebuf, codepage_path);
	strlcat(linebuf, PATHSEP "translit", sizeof(linebuf));

	if ((in = fopen(linebuf, "r")) == NULL)
		return NULL;

	size = 0;
	ucmin = ~0;
	ucmax = 0;

	/* pass one - calculate the size of the table */
	while ((str = read_line(linebuf, sizeof(linebuf), in)) != NULL) {
		u_int32_t field;

		if (sscanf(str, "0x%4x", &field) != 1)
			continue;

		uc = (u_short)field;

		for (p = str + 2; isxdigit(*p); ++p);
		for (; IsWS(*p); ++p);

		if ((sz = strlen(p)) == 0)
			continue;

		if (uc < ucmin)
			ucmin = uc;
		if (uc > ucmax)
			ucmax = uc;

		size += sz + 1;
	}

	if ((cs = calloc(1, sizeof(TransLitSet) + (ucmax - ucmin + 1)
	* sizeof(char*))) == NULL) {
		fclose(in);
		return NULL;
	}

	if ((map = calloc(1, size)) == NULL) {
		fclose(in);
		free(cs);
		return NULL;
	}

	cs->map = map;
	cs->ucmin = ucmin;
	cs->ucmax = ucmax;

	/* second step, this time retreiving the data */
	fseek(in, 0L, SEEK_SET);
	while ((str = read_line(linebuf, sizeof(linebuf), in)) != NULL) {
		u_int32_t field;

		if (sscanf(str, "0x%4x", &field) != 1)
			continue;

		uc = (u_short)field;

		for (p = str + 2; isxdigit(*p); ++p);
		for (; IsWS(*p); ++p);

		if ((sz = strlen(p)) == 0)
			continue;

		strcpy(map, p);
		cs->uc2seq[uc - ucmin] = map;
		map += sz + 1;
	}
	fclose(in);

	/* map unknown sequences */
	for (uc = ucmin; uc <= ucmax; ++uc)
		if (cs->uc2seq[uc - ucmin] == NULL)
			cs->uc2seq[uc - ucmin] = "?";

	return cs;
}


static CharSet *read_enc(const char *fname)
{
	FILE *in;
	unsigned char ac, linebuf[1024], *p;
	unsigned short a2u[256], uc, ucmin, ucmax;
	size_t size;
	CharSet *cs;
int lines = 0;

	/* assuming that we never pass a const string here */
	if (p = strchr(fname, '\\'))
		*p = '\0';
	if (p = strchr(fname, '/'))
		*p = '\0';

	strcpy(linebuf, codepage_path);
	strlcat(linebuf, PATHSEP, sizeof(linebuf));
	strlcat(linebuf, fname, sizeof(linebuf));

	if ((in = fopen(linebuf, "r")) == NULL)
		return NULL;

	ucmin = ~0;
	ucmax = 0;

	memset(a2u, 0, sizeof(a2u));

	while ((p = read_line(linebuf, sizeof(linebuf), in)) != NULL) {
		u_int32_t field;

		if (sscanf(p, "0x%2x", &field) != 1)
			continue;

		ac = (u_char)field;

		for (p += 2; isxdigit(*p); ++p);
		for (; IsWS(*p); *p++ = '\0');

		if (sscanf(p, "0x%4x", &field) != 1)
			continue;

		uc = (u_short)field;
		if (uc > ucmax)
			ucmax = uc;
		if (uc < ucmin)
			ucmin = uc;

		a2u[ac] = uc;
	}

	fclose(in);

	size = sizeof(CharSet) + (ucmax - ucmin) * 2;
	if ((cs = calloc(1, size)) == NULL)
		return NULL;

	cs->name = strdup(fname);
	cs->ucmin = ucmin;
	cs->ucmax = ucmax;

	memcpy(cs->a2u, a2u, sizeof(cs->a2u));
	for (size = 0; size < 256; ++size)
		cs->u2a[a2u[size] - ucmin] = size;

	return cs;
}


/* returns pointer to a charset with the given name, increasing its usage count,
 * returns NULL if codepage with the given name could not be found.
 */
CharSet *charset_attach(const char *name)
{
	CharSet *cs, *last;
	CharSetAlias *ca;

	for (cs = charsets, last = cs; cs; cs = cs->next) {
		if (strcasecmp(name, cs->name) == 0) {
			++cs->refcount;
			return cs;
		}

		for (ca = cs->alias; ca; ca = ca->next) {
			if (strcasecmp(name, ca->name) == 0) {
				++cs->refcount;
				return cs;
			}
		}

		last = cs;
	}

	if ((cs = read_enc(name)) == NULL)
		return NULL;

	if (last == NULL)
		charsets = cs;
	else {
		cs->prev = last;
		last->next = cs;
	}

	++cs->refcount;
	return cs;
}


void charset_detach(CharSet *cs)
{
	--cs->refcount;
}


/* releases unused charsets, refreshing the loaded ones when possible */
void charset_refresh(void)
{
	CharSet *cs, *tmp;
	TransLitSet *tl;

	for (cs = charsets; cs; cs = tmp) {
		if (cs->refcount == 0) {
			tmp = cs->next;
			if (cs->prev != NULL)
				cs->prev->next = cs->next;
			if (cs->next != NULL)
				cs->next->prev = cs->prev;
			free(cs);
		}
		else if ((tmp = read_enc(cs->name)) != NULL) {
			int i;
			aClient *acptr;

			tmp->alias = cs->alias;
			tmp->next = cs->next;
			tmp->prev = cs->prev;
			tmp->refcount = cs->refcount;
			free(cs);

			for (i = highest_fd; i >= 0; --i) {
				if ((acptr = local[i]) && IsPerson(acptr)) {
					if (acptr->charset == cs)
						acptr->charset = tmp;
				}
			}

			tmp = tmp->next;
		}
		/* this only happens when we fail to load a codepage that has
		 * a number of references, we can't just release it - leave it
		 * untouched.
		 */
		else tmp = cs->next;

		if (cs == charsets)
			charsets = tmp;
	}

	if ((tl = read_translit()) != NULL) {
		if (cs_translit != NULL) {
			free(cs_translit->map);
			free(cs_translit);
		}
		cs_translit = tl;
	}
}


/* converts a string from a specific charset to UTF8.  the string might
 * grow a bit, so the buffer must be large enough.
 */
void charset_convertfrom(char *buf, size_t size, CharSet *cs)
{
	int changed;
	char tmpbuf[2048], *src, *dst, *srcl, *dstl, *p;
	unsigned char ac;
	unsigned short uc;

	src = buf;
	srcl = buf + size;
	dst = tmpbuf;
	dstl = tmpbuf + sizeof(tmpbuf);

	for (; *src != '\0' && src < srcl && dst < dstl; ++src) {
		ac = *src;

		if (ac & 0x80 == 0x00) {
			*dst++ = ac;
			continue;
		}

		changed = 1;

		if ((uc = cs->a2u[ac]) == 0x00)
			*dst++ = '?';
		else if ((uc & 0xff80) == 0x00)
			*dst++ = ac;
		else
			for (p = uc2utf(uc); *p != '\0'; *dst++ = *p++);
	}

	if (changed == 0)
		return;

	if (dst == dstl)
		--dst;
	*dst = '\0';

	strlcpy(buf, tmpbuf, size);
}


/* converts an UTF8 string to the given charset, applying transliteration.
 * the does not grow under normal circumstances, relax.
 */
void charset_convertto(char *buf, size_t size, CharSet *cs)
{
	int changed;
	char tmpbuf[2048], *src, *dst, *srcl, *dstl, *p;
	unsigned char c;
	unsigned short uc;

	changed = 0;
	src = buf;
	srcl = buf + size;
	dst = tmpbuf;
	dstl = tmpbuf + sizeof(tmpbuf);

	/* two almost similar parts to optimize checks a bit */

	if (cs_translit == NULL) {
		while (*src != '\0' && src < srcl && dst < dstl) {
			if (*src & 0x80 == 0) {
				*dst++ = *src++;
				continue;
			}

			changed = 1;

			uc = utf2uc((unsigned char **)&src);

			if ((uc >= cs->ucmin) && (uc <= cs->ucmax)
			&& (c = cs->u2a[uc - cs->ucmin]) != 0)
				*dst++ = c;
			else
				*dst++ = '?';
		}
	}

	/* we also have a transliteration map */
	else {
		while (*src != '\0' && src < srcl && dst < dstl) {
			if (*src & 0x80 == 0) {
				*dst++ = *src++;
				continue;
			}

			changed = 1;

			uc = utf2uc((unsigned char **)&src);

			if ((uc >= cs->ucmin) && (uc <= cs->ucmax)
			&& (c = cs->u2a[uc - cs->ucmin]) != 0)
				*dst++ = c;

			/* notice that we don't check for null pointers, the
			 * map must be prepared so that the empty sequences
			 * are replaces with a dummy string (and it is so)
			 */
			else if ((uc >= cs_translit->ucmin)
			&& (uc <= cs_translit->ucmax)) {
				p = cs_translit->uc2seq[uc
					- cs_translit->ucmin];
				for (; *p != '\0'; *dst++ = *p++);
			}
			else
				*dst++ = '?';
		}
	}

	if (changed == 0)
		return;

	/* truncate the new string */
	if (dst == dstl)
		--dst;
	*dst = '\0';

	strlcpy(buf, tmpbuf, size);
}
